/* 
 * program do rysowania wykresow dowolnych funkcji
 *
 * Autor: Marek Szyprowski <marek@amiga.pl>
 *
 * PlotXY v1.3
 *
 * Program PlotXY jest wolnym oprogramowaniem; moesz go
 * rozprowadza dalej i/lub modyfikowa na warunkach Powszechnej
 * Licencji Publicznej GNU, wydanej przez Fundacj Wolnego
 * Oprogramowania - wedug wersji 2 tej Licencji lub (wedug twojego
 * wyboru) ktrej z pniejszych wersji.
 *
 * Niniejszy program rozpowszechniany jest z nadziej, i bdzie on
 * uyteczny - jednak BEZ JAKIEJKOLWIEK GWARANCJI, nawet domylnej
 * gwarancji PRZYDATNOCI HANDLOWEJ albo PRZYDATNOCI DO OKRELONYCH
 * ZASTOSOWA. W celu uzyskania bliszych informacji signij do
 * Powszechnej Licencji Publicznej GNU.
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <SDL/SDL.h>
#include "SDL_font.h"
#include "rpn.h"
#include "line.h"

#define VERSION "PlotXY v1.3 (02.05.2003)  (C) by Marek Szyprowski <march@staszic.waw.pl>"
#define FUNKCJA "x^3 - 3*x - 20"

#define WIDTH 800
#define HEIGHT 600

#define TRUE 1
#define FALSE 0

#define put_pixel(x, y, c) (*(pixels+(int)x+WIDTH*(int)y)=c)
#define line(x1, y1, x2, y2, c) bresline(x1, y1, x2, y2, c, WIDTH)

/* *** */
double var_x;
char *funkcja = FUNKCJA;
Token funkcja_t[0xff];

double f (double x)
{
	var_x = x;
	return eval_rpn(funkcja_t);
}

double g (double x)
{
	return (f(x+0.000001)-f(x))/0.000001;
}
/* *** */

SDL_Surface *screen;
unsigned char *pixels;
SDL_Font *font;

/****/
void init_sdlgfx (void)
{
	if((SDL_Init(SDL_INIT_VIDEO)==-1))
	{
		printf("Could not initialize SDL: %s.\n", SDL_GetError());
		exit(-1);
	}
	if ((screen = SDL_SetVideoMode(WIDTH, HEIGHT, 8, SDL_FULLSCREEN)) == NULL)
	{
		printf("Couldn't set up video mode: %s\n", SDL_GetError());
		exit(-1);
	}
	pixels = screen->pixels;
	font = SDL_LoadFont(screen, NULL);
	SDL_WarpMouse(WIDTH-1, HEIGHT-1);
}

#define sx(x)  (int)((x-xmin)*k)
#define sy(y) (HEIGHT - (int)((y-ymin)*k))

int mod(double a, double b)
{
	int i;
	if (a > 0)
		for (i=0; (i)*b < a; i++)
			;
	else
		for (i=0; (i-1)*b > a; i--)
			;
	return i;
}

void legenda(int pos)
{
	int x, y;
	if (pos > 0)
	{
		switch (pos)
		{
		case 1:
			x = 0; y = 3.5*font->height;
			break;
		case 2: 
			x = 0; y = HEIGHT-font->height*8;
			break;
		case 3:
			x = WIDTH- font->width*20; y = 3.5*font->height;
			break;
		default: 
			x = WIDTH- font->width*20; y = HEIGHT-font->height*8;
		}
		SDL_PrintF(screen, x, y, font, " -=* Legenda *=-\nA/Z - dokadno\n+/- - powikszenie\nstrzaki - pozycja\nT - metoda rysowania\nL - legenda\nP - pochodna funkcji\nQ - wyjcie");	 
		SDL_PrintF(screen, 0, 0, font, VERSION);
		SDL_PrintF(screen, 0, font->height+1, font, "Program dostpny na zasadach Powszechnej Licencji Publicznej GNU.\nWicej informacji w pliku \"GPL.txt\".");
	}
}

void osie (double k, double d, double xmin, double ymin)
{
	double skala=1;
	char *format;
	
	if (k > 64)
	{
		while(skala*k > 64)
			skala*=0.5;
		skala*=2;		
	}		
	else
		while (skala*k < 64)
			skala*=2;
	
	if (skala > 10)
		format = "%.0f";
	else if (skala > 1)
		format = "%.1f";
	else if (skala > 0.1)
		format = "%.2f";
	else if (skala > 0.01)
		format = "%.3f";
	else if (skala > 0.001)
		format = "%.4f";
	else if (skala > 0.0001)
		format = "%.5f";
	else
		format = "%f";

	if (sy(0)-2 >= 0 && sy(0)+2 < HEIGHT)
	{
		double tmp;
		tmp = mod(xmin,skala) * skala;
		line(0, sy(0), WIDTH-1, sy(0), 16);
		while (sx(tmp) < WIDTH-1)
		{
			line(sx(tmp), sy(0)-2, sx(tmp), sy(0)+2, 16);
			SDL_PrintF(screen, sx(tmp)+2, sy(0)+2, font, format, tmp);
			tmp+=skala;
		}
	}	
	if (sx(0)-2 >= 0 && sx(0)+2 < WIDTH)
	{
		double tmp;
		tmp = mod(ymin,skala) * skala;
		line(sx(0), 0, sx(0), HEIGHT-1, 16);
		while (sy(tmp) > 0)
		{
			if (sy(tmp) < HEIGHT)
			{
				line(sx(0)-2, sy(tmp), sx(0)+2, sy(tmp), 16);
				SDL_PrintF(screen, sx(0)+2, sy(tmp)+2, font, format, tmp);
			}
			tmp+=skala;
		}
	}
}

void plot_points (double k, double d, double xmin, double ymin, int c, double (*calc)(double))
{
	double x, y;

	for (x=xmin; sx(x) < WIDTH; x+=d)
	{
		y = (*calc)(x);
		if (sy(y) > 0 && sy(y) < HEIGHT)
			put_pixel(sx(x), sy(y), c);
	}
}
void plot_lines (double k, double d, double xmin, double ymin, int c, double (*calc)(double))
{
	double xn, xs, yn, ys;
	xs = xmin;
	ys = (*calc)(xs);
	for (xn=xmin; sx(xn) < WIDTH; xn+=d)
	{
		yn = (*calc)(xn);
		if (sy(yn) > 0 && sy(yn) < HEIGHT && sy(ys) > 0 && sy(ys) < HEIGHT)
			line(sx(xn), sy(yn), sx(xs), sy(ys), c);
		xs = xn;
		ys = yn;
	}
}

void plot_f(double k, double d, double xmin, double ymin, int l, int t, int p)
{
	void (*plot_type)(double, double, double, double, int, double (*calc)(double));
	plot_type = (t) ? plot_points : plot_lines;

	SDL_FillRect(screen, 0, 0);

	legenda(l);
	osie(k, d, xmin, ymin);

	SDL_UpdateRect(screen, 0, 0, 0, 0);

	if (p == 0 || p == 1)
		(*plot_type)(k, d, xmin, ymin, 10, f);

	SDL_UpdateRect(screen, 0, 0, 0, 0);
	
	if (p == 1 || p == 2)
		(*plot_type)(k, d, xmin, ymin, 80, g);

	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

/****/

#define ZOOMUP 1.2
#define ZOOMDOWN 0.8

void plot (void)
{
	SDL_Event ev;
	int q=0;
	double k = 10; 
	double d = 0.2;
	double xmin = -WIDTH/2/k;
	double ymin = -HEIGHT/2/k;
	double tmpx, tmpy, tmpd;
	int l=1, t=0, p=0;
	
	int kup = FALSE, kdown = FALSE, kleft = FALSE, kright = FALSE;
	int changed = TRUE;
	
	init_sdlgfx();
	
	while (!q)
	{
	    while (SDL_PollEvent(&ev))
		{
			if (ev.type == SDL_KEYDOWN)
			{	switch (ev.key.keysym.sym)
				{
					case SDLK_q:
					case SDLK_ESCAPE:
						q++;
						break;
					case SDLK_EQUALS:
						tmpx = xmin + WIDTH/2/k;
						tmpy = ymin + HEIGHT/2/k;
						tmpd = k*d;
						k*=ZOOMUP;
						xmin = tmpx - WIDTH/2/k;
						ymin = tmpy - HEIGHT/2/k;
						d = tmpd/k;
						changed = TRUE;
						break;
					case SDLK_MINUS:
						tmpx = xmin + WIDTH/2/k;
						tmpy = ymin + HEIGHT/2/k;
						tmpd = k*d;
						k*=ZOOMDOWN;
						xmin = tmpx - WIDTH/2/k;
						ymin = tmpy - HEIGHT/2/k;
						d = tmpd/k;
						changed = TRUE;
						break;
					case SDLK_a:
						d/=2;
						changed = TRUE;
						break;
					case SDLK_z:
						d*=2;
						changed = TRUE;
						break;
					case SDLK_l:
						if (l < 4)
							l++;
						else
							l = 0;
						changed = TRUE;
						break;
					case SDLK_t:
						t ^= 1;
						changed = TRUE;
						break;
					case SDLK_p:
						if (p<2) p++; else p=0;
						changed = TRUE;
						break;
					case SDLK_UP:
						kup = TRUE;
						break;
					case SDLK_DOWN:
						kdown = TRUE;
						break;
					case SDLK_RIGHT:
						kright = TRUE;
						break;
					case SDLK_LEFT:
						kleft = TRUE;
						break;
					default:
						break;
				}
			}
			else if (ev.type == SDL_KEYUP)
			{	switch (ev.key.keysym.sym)
				{
					case SDLK_UP:
						kup = FALSE;
						break;
					case SDLK_DOWN:
						kdown = FALSE;
						break;
					case SDLK_RIGHT:
						kright = FALSE;
						break;
					case SDLK_LEFT:
						kleft = FALSE;
						break;
					default:
						break;
				}
			}
		}
		
		SDL_Delay(25);
		
		if (kup)
		{
			ymin += 10/k;
			changed = TRUE;
		}
		if (kdown)
		{
			ymin -= 10/k;
			changed = TRUE;
		}
		if (kright)
		{
			xmin += 10/k;
			changed = TRUE;
		}
		if (kleft)
		{
			xmin -= 10/k;
			changed = TRUE;
		}

		if (changed)
		{
			plot_f(k, d, xmin, ymin, l, t, p);
			changed = FALSE;
		}
	}
									 
	SDL_FreeFont(font);
	SDL_Quit();
}
 
/* */
char *getline (void)
{
	static char line[0xff];
	char *s = line;
	int c;
	while ((c = getchar()) != '\n' && c != EOF)
	*s++ = c;
	*s = '\0';
	return line;
}
/* */

int main (int argc, char *argv[])
{
	char *p, *res;

	printf(VERSION "\nProgram rysuje wykres dowolnej funkcji\n\n"
	"Program ten jest wolnym oprogramowaniem; moesz go\n"
	"rozprowadza dalej i/lub modyfikowa na warunkach Powszechnej\n"
	"Licencji Publicznej GNU. Niniejszy program rozpowszechniany jest\n" 
	"z nadziej, i bdzie on uyteczny - jednak BEZ JAKIEJKOLWIEK\n"
	"GWARANCJI. W celu uzyskania bliszych informacji signij do\n"
	"Powszechnej Licencji Publicznej GNU (plik gpl.txt).\n\n");

	if (argc == 2)
		funkcja = argv[1];
	else
	{
		printf("Wprowad wzr funkcji (ENTER -> \"f(x) = %s\")\nf(x) = ", funkcja);
		p = getline();
		if (*p != '\0')
			funkcja = p;
	}
	res = str2rpn(funkcja, funkcja_t);
	
	if (res != NULL)
	{
		printf("Wystpi bd! %s\n", res);
		return 5;
	}

	printf("Wprowadzono funkcj:\nf(x) = %s\n\n", funkcja);

	plot();

	return 0;
}
