/*
 * Funkcje zamieniajce zapis tradycyjny na odwrotn notacj polsk
 *
 *   rpn.c v1.3 (17.02.2003) by Marek Szyprowski <marek@amiga.pl>
 *
 * Niniejszy fragment kodu jest czci biblioteki RPN 
 *
 
 * Biblioteka RPN jest wolnym oprogramowaniem; moesz go
 * rozprowadza dalej i/lub modyfikowa na warunkach Powszechnej
 * Licencji Publicznej GNU, wydanej przez Fundacj Wolnego
 * Oprogramowania - wedug wersji 2 tej Licencji lub (wedug twojego
 * wyboru) ktrej z pniejszych wersji.
 *
 * Niniejszy program rozpowszechniany jest z nadziej, i bdzie on
 * uyteczny - jednak BEZ JAKIEJKOLWIEK GWARANCJI, nawet domylnej
 * gwarancji PRZYDATNOCI HANDLOWEJ albo PRZYDATNOCI DO OKRELONYCH
 * ZASTOSOWA. W celu uzyskania bliszych informacji signij do
 * Powszechnej Licencji Publicznej GNU.
 *	   
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "rpn.h"
#include "eval.h"

extern struct op operators[];

#define stack_put(t) (*stack_p++ = (t))
#define stack_get() (*--stack_p)
#define stack_check() (*(stack_p-1))

int checkpar_str (char *str)
{
	int count=0;

	while (*str != '\0')
	{
		if (*str == '(')
			count++;
		else if (*str == ')')
			count--;
		str++;
	}
	if (count == 0)
		return OK;
	else
		return ERROR;
}

int check_unk (struct token *data)
{
	while (data->type != End)
	{
		if (data->type == Unknown)
			return ERROR;
		data++;
	}
	return OK;
}

int check_rpn (struct token *data)
{
	int stack = 0;

	while (data->type != End)
	{
		if (data->type == Number || data->type == Variable)
			stack++;
		else if (data->type == Operator && data->val.op->args == 2)
			stack--;
		data++;
	}

	if (stack == 1)
		return OK;
	else
		return ERROR;
}

void std2rpn (struct token *src, struct token *dst)
{
	static struct token stack[CONV_STACK_LEN];
	static struct token *stack_p=stack;
	struct token tmp;

	tmp.type = End;
	stack_put(tmp);
	 
	while (src->type != End)
	{
		if (src->type == Number || src->type == Variable)
			*dst++ = *src;
		else if (src->type == Lparenth)
			stack_put(*src);
		else if (src->type == Rparenth)
			while ((tmp = stack_get()).type != Lparenth)
				*dst++ = tmp;
		else if (src->type == Operator)
		{
			while ((stack_check()).type != End && (stack_check()).type != Lparenth && src->val.op->pri <= (stack_check()).val.op->pri)
				*dst++ = stack_get();
			stack_put(*src);
		}
		src++;
	}
	do {
		tmp = stack_get();
		*dst++ = tmp;
	} while (tmp.type != End);

}

double eval_rpn(struct token *data)
{
	static double stack[EVAL_STACK_LEN];
	double *stack_p=stack;

	while (data->type != End)
	{
		if (data->type == Number)
			stack_put(data->val.val);
		else if (data->type == Operator && data->val.op->args == 2)
		{
			double a, b;
			a = stack_get();
			b = stack_get();
			stack_put((*data->val.op->eval)(b, a));
		}
		else if (data->type == Operator && data->val.op->args == 1)
		{
			double a;
			a = stack_get();
			stack_put((*data->val.op->eval)(a, 0));
		}
		else if (data->type == Variable)
			stack_put((*data->val.op->eval)(0, 0));
		data++;
	}
	return stack_get();
}

void str2tokens (char *str, struct token *ta)
{
	char *p;
	struct token *t = ta;
	struct op *mul_op;

	for (mul_op = operators; mul_op->symbol != NULL; mul_op++)
		if (mul_op->symbol[0] == '*')
			break; 

	p = str;
	while (*p != '\0')
	{
		if (isspace(*p))
			p++;
		else if (*p == '(')
		{
			t->type = Lparenth;
			p++; t++;
		}
		else if (*p == ')')
		{
			t->type = Rparenth;
			p++; t++;
		}
		else if (isdigit(*p) || *p == '.')
		{
			t->val.val = strtod(p, &p);
			t->type = Number;
			t++;	
		}
		else
		{
			struct op *tmpop;

			for (tmpop = operators; tmpop->symbol != NULL; tmpop++)
				if (strncmp(p, tmpop->symbol, strlen(tmpop->symbol)) == 0)
					break;

			t->val.op = tmpop;
			if (tmpop->symbol == NULL)
			{
				t->type = Unknown;
				p++;
			}
			else
			{
				p += strlen(tmpop->symbol);
				if (tmpop->args == 0)
					t->type = Variable;
				else
					t->type = Operator;
			}
			/* odpowiednie potraktowanie minusa */
			if (t->type == Operator && t->val.op->symbol[0] == '-' && (t == ta || ((t-1)->type == Operator || (t-1)->type == Lparenth)))
			{
				t->type = Number;
				t->val.val = -1.0;
				t++;
				t->type = Operator;
				t->val.op = mul_op;
			}
			t++;
		}
	}
	t->type = End;
}

char *str2rpn(char *src, struct token *dst)
{
	struct token tmp[TEMP_ARRAY_LEN];
	 
	if (checkpar_str(src) == ERROR)
		return "Nie zamknity nawias.";

	str2tokens(src, tmp);

	if (check_unk(tmp) == ERROR)
		return "We wzorze znajduje si nierozpoznany symbol.";

	std2rpn(tmp, dst);

	if (check_rpn(dst) == ERROR)
		return "Brak argumentw dla wszystkich operatorw.";

	return NULL;
}
